import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';

import '../plugin_annotations.dart';

class _StartupTaskGenerator extends Generator {
  static const startupTask = TypeChecker.fromRuntime(StartupTask);
  static const onekitMixin = TypeChecker.fromRuntime(OneKitMixin);
  @override
  String generate(LibraryReader library, BuildStep buildStep) {
    String concat(String prev, String ele) => prev + ele;
    final allLibs = <LibraryElement>{};
    final annotatedLibs = <LibraryElement>[];
    final workingList = [library.element];
    while (workingList.isNotEmpty) {
      final lib = workingList.removeLast();
      if (allLibs.contains(lib)) {
        continue;
      }
      allLibs.add(lib);
      workingList
        ..addAll(lib.importedLibraries
            .where((transLib) => !allLibs.contains(transLib)))
        ..addAll(lib.exportedLibraries
            .where((transLib) => !allLibs.contains(transLib)));
      if (lib.importedLibraries
          .where((transLib) =>
              '/ve_code_gen/lib/plugin_annotations.dart' ==
              '${transLib.librarySource}')
          .isNotEmpty) {
        annotatedLibs.add(lib);
      }
    }
    // Quick reject
    if (annotatedLibs.isEmpty) {
      return '''
import 'package:flutter/widgets.dart';
import 'package:ve_onekit/ve_onekit.dart';

void runOneKitApp(Function builder) {
  WidgetsFlutterBinding.ensureInitialized();
  OneKit.runApp(builder);
}
''';
    }

    final indexedAnnotatedLibs = annotatedLibs.asMap();
    final imports = indexedAnnotatedLibs
        .map((ind, lib) {
          final splitedPath = lib.librarySource.toString().split('/');
          final file = splitedPath.sublist(3).join('/');
          final pack = splitedPath[1];
          return MapEntry(ind, "import 'package:$pack/$file' as _i$ind;\n");
        })
        .values
        .fold('', concat);
    final tasks = indexedAnnotatedLibs.map((ind, lib) =>
        MapEntry(ind, LibraryReader(lib).annotatedWith(startupTask)));
    final mixins = indexedAnnotatedLibs.map((ind, lib) =>
        MapEntry(ind, LibraryReader(lib).annotatedWith(onekitMixin)));
    final preTasks = _getTasksString(tasks, 'high');
    final preTaskString = preTasks.isEmpty
        ? ''
        : '''
final _preTaskList = <Function>[
  $preTasks
];''';
    final normalTasks = _getTasksString(tasks, 'normal');
    final normalTaskString = normalTasks.isEmpty
        ? ''
        : '''
final _taskList = <Function>[
  $normalTasks
];
''';
    final postTasks = _getTasksString(tasks, 'low');
    final postTasksString = postTasks.isEmpty
        ? ''
        : '''
final _postTaskList = <Function>[
  $postTasks
];
''';
    final mix = _getMixinString(mixins);
    final mixinsString = mix.isEmpty ? '' : 'with $mix';
    final res = '''
import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:ve_onekit/ve_onekit.dart';
$imports

$preTaskString

$normalTaskString

$postTasksString

void runOneKitApp(Function builder) async {
  ${preTasks.isNotEmpty ? 'await Future.forEach<Function>(_preTaskList, (f) => f());' : ''}
  OneKit.runApp(([_]) async {
    OneKitWidgetsBinding.ensureInitialized();
    ${normalTasks.isNotEmpty ? 'await Future.forEach<Function>(_taskList, (f) => f());' : ''}
    ${postTasks.isNotEmpty ? 'await Future.forEach<Function>(_postTaskList, (f) => f());' : ''}
    final FutureOr<Widget> widget = _ == null ? builder() : builder(_);
    return widget;
  });
}

class OneKitWidgetsBinding extends WidgetsFlutterBinding $mixinsString {
  static WidgetsBinding? ensureInitialized() {
    if (WidgetsBinding.instance == null) {
      OneKitWidgetsBinding();
    }
    return WidgetsBinding.instance;
  }
}
''';
    return res;
  }

  String _getTasksString(
      Map<int, Iterable<AnnotatedElement>> tasks, String priority) {
    final taskCollection = {};
    tasks.forEach((ind, eles) => eles
            .where((ele) =>
                ele.annotation.read('_priority').isString &&
                ele.annotation.read('_priority').stringValue == priority)
            .forEach((ele) {
          final name = ele.element.name ?? '';
          if (name.isEmpty) {
            return;
          }
          if (name.endsWith('Pro')) {
            taskCollection[name.substring(0, name.length - 3)] =
                '_i$ind.${ele.element.name}.init';
          } else if (!taskCollection.containsKey(name)) {
            taskCollection[name] = '_i$ind.${ele.element.name}.init';
          }
        }));

    return taskCollection.values.join(',');
  }

  String _getMixinString(Map<int, Iterable<AnnotatedElement>> mixins) {
    final mixinCollection = {};
    mixins.forEach((ind, eles) => eles.forEach((ele) {
          final name = ele.element.name ?? '';
          if (name.isEmpty) {
            return;
          }
          if (name.endsWith('Pro')) {
            mixinCollection[name.substring(0, name.length - 3)] =
                '_i$ind.${ele.element.name}';
          } else if (!mixinCollection.containsKey(name)) {
            mixinCollection[name] = '_i$ind.${ele.element.name}';
          }
        }));

    return mixinCollection.values.join(',');
  }
}

Builder startupTaskGeneratorBuilder(BuilderOptions options) =>
    LibraryBuilder(_StartupTaskGenerator(),
        generatedExtension: '.startup.dart');
